import hashlib
import os
import sys
import traceback
import warnings
from collections import defaultdict
from configparser import ConfigParser

import fsui
from fsbc.application import app
from fsbc.settings import Settings
from fsbc.task import Task
from fsbc.util import is_uuid, unused
from fsgs.amiga import whdload
from fsgs.amiga.amiga import Amiga
from fsgs.application import ApplicationMixin
from fsgs.archive import Archive
from fsgs.context import fsgs
from fsgs.download import Downloader
from fsgs.filedatabase import FileDatabase
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.input.enumeratehelper import EnumerateHelper
from fsgs.platform import PLATFORM_IDS, PlatformHandler
from fsgs.util.archiveutil import ArchiveUtil
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.option import Option
from launcher.startup_scan import StartupScan
from launcher.ui.behaviors.platformbehavior import AMIGA_PLATFORMS
from launcher.ui.download import DownloadGameWindow, DownloadTermsDialog
from launcher.ui.launch import LaunchDialog
from launcher.ui.launcherwindow import LauncherWindow


class StringDict(defaultdict):
    def get(self, item, default=""):
        return super().get(item, default)


class LauncherApp(ApplicationMixin, fsui.Application):
    auto_detect_game = True

    def __init__(self):
        fsui.Application.__init__(self, "fs-uae-launcher")
        self.set_icon(fsui.Icon("fs-uae-launcher", "pkg:launcher"))
        self.fsgs = fsgs
        # Try to auto-detect game when launching with an archive

    def on_idle(self):
        self.fsgs.signal.process()

    @classmethod
    def pre_start(cls):
        print("FSUAELauncherApplication.pre_start")
        cls.load_settings()
        StartupScan.config_startup_scan()
        StartupScan.kickstart_startup_scan()
        # FIXME: should now sanitize check some options -for instance,
        # - check if configured joysticks are still connected
        # - check if paths still exists, etc
        # Config.update_kickstart()
        cls.load_plugins()

    @classmethod
    def start(cls):
        cls.pre_start()
        print("FSUAELauncherApplication.start")

        if cls.run_config_or_game():
            return False
        # window = LauncherWindow(fsgs)
        # window.show()
        LauncherWindow.open()
        return True

    @classmethod
    def run_config_or_game(cls):
        config_path = None
        archive_path = None
        floppy_image_paths = []
        cdrom_image_paths = []
        config_uuid = None
        floppy_extensions = (".adf", ".ipf", ".dms", ".adz")
        cdrom_extensions = (".cue", ".iso")
        archive_extensions = (".zip", ".lha")

        # FIXME: replace argument "parsing" with use of argparse module
        # at some point

        last_arg = sys.argv[-1]
        file_ext = os.path.splitext(last_arg)[-1].lower()
        if file_ext == ".fs-uae":
            config_path = last_arg
        elif file_ext in archive_extensions:
            archive_path = last_arg
        # elif file_ext in floppy_extensions:
        #     floppy_image_paths = [last_arg]
        elif is_uuid(last_arg):
            config_uuid = last_arg.lower()
        for arg in sys.argv[1:]:
            if not arg.startswith("--"):
                _, ext = os.path.splitext(arg)
                if ext in floppy_extensions:
                    floppy_image_paths.append(arg)
                elif ext in cdrom_extensions:
                    cdrom_image_paths.append(arg)

        if config_path:
            print("[STARTUP] Config path given:", config_path)
            if not os.path.exists(config_path):
                print("[STARTUP] Config path does not exist", file=sys.stderr)
                return True
            LauncherConfig.load_file(config_path)
            fsgs.config.add_from_argv()
            return cls.run_config_directly()

        if archive_path:
            print("[STARTUP] Archive path given:", archive_path)
            if not os.path.exists(archive_path):
                print("[STARTUP] Archive path does not exist", file=sys.stderr)
                return True
            archive = Archive(os.path.realpath(archive_path))
            archive_name = os.path.basename(archive_path)
            # We want to exclude pure directory entries when checking for
            # archives with only floppies.
            arc_files = [
                x for x in archive.list_files() if not x.endswith("/")
            ]
            if all(
                map(lambda f: f.lower().endswith(floppy_extensions), arc_files)
            ):
                print("[STARTUP] Archive contains floppy disk images only")
                floppy_image_paths = arc_files
            else:
                if cls.auto_detect_game:
                    # FIXME: Could also do this for floppy file archives.
                    archive_util = ArchiveUtil(archive_path)
                    archive_uuid = archive_util.create_variant_uuid()
                    print(
                        "[STARTUP] Try auto-detecting variant, uuid =",
                        archive_uuid,
                    )
                    if fsgs.load_game_variant(archive_uuid):
                        print("[STARTUP] Auto-detected variant", archive_uuid)
                        print("[STARTUP] Adding archive files to file index")
                        for archive_file in archive.list_files():
                            stream = archive.open(archive_file)
                            data = stream.read()
                            size = len(data)
                            sha1 = hashlib.sha1(data).hexdigest()
                            FileDatabase.add_static_file(
                                archive_file, size=size, sha1=sha1
                            )
                        fsgs.config.add_from_argv()
                        fsgs.config.set("__config_name", archive_name)
                        LauncherConfig.post_load_values(fsgs.config)
                        return cls.run_config_directly()

                values = whdload.generate_config_for_archive(archive_path)
                values["hard_drive_0"] = archive_path
                values.update(fsgs.config.config_from_argv())
                # archive_name, archive_ext = os.path.splitext(archive_name)
                values["__config_name"] = archive_name
                return cls.run_config_directly_with_values(values)

        if floppy_image_paths:
            enum_paths = tuple(enumerate(floppy_image_paths))
            values = {}
            values.update(fsgs.config.config_from_argv())
            max_drives = int(values.get("floppy_drive_count", "4"))
            values.update(
                {
                    "floppy_drive_{0}".format(k): v
                    for k, v in enum_paths[:max_drives]
                }
            )
            values.update(
                {"floppy_image_{0}".format(k): v for k, v in enum_paths[:20]}
            )
            # FIXME: Generate a better config name for save dir?
            values["__config_name"] = "Default"
            return cls.run_config_directly_with_values(values)

        if cdrom_image_paths:
            enum_paths = tuple(enumerate(cdrom_image_paths))
            values = {"amiga_model": "CD32"}
            values.update(fsgs.config.config_from_argv())
            max_drives = int(values.get("cdrom_drive_count", "1"))
            values.update(
                {
                    "cdrom_drive_{0}".format(k): v
                    for k, v in enum_paths[:max_drives]
                }
            )
            values.update(
                {"cdrom_image_{0}".format(k): v for k, v in enum_paths[:20]}
            )
            # FIXME: Generate a better config name for save dir?
            values["__config_name"] = "Default"
            return cls.run_config_directly_with_values(values)

        if config_uuid:
            print("[STARTUP] Config uuid given:", config_uuid)
            variant_uuid = config_uuid
            # values = fsgs.game.set_from_variant_uuid(variant_uuid)
            if fsgs.load_game_variant(variant_uuid):
                print("[STARTUP] Loaded variant")
            else:
                print("[STARTUP] Could not load variant, try to load game")
                game_uuid = config_uuid
                variant_uuid = fsgs.find_preferred_game_variant(game_uuid)
                print("[STARTUP] Preferred variant:", variant_uuid)
                fsgs.load_game_variant(variant_uuid)
            fsgs.config.add_from_argv()
            LauncherConfig.post_load_values(fsgs.config)
            return cls.run_config_directly()

    @classmethod
    def run_config_directly(cls):
        cls.start_game()
        return True

    @classmethod
    def run_config_directly_with_values(cls, values):
        LauncherConfig.load(values)
        LauncherConfig.post_load_values(values)
        return cls.run_config_directly()

    _plugins_loaded = False

    @classmethod
    def load_plugins(cls):
        if cls._plugins_loaded:
            return
        cls._plugins_loaded = True

        plugins_dir = FSGSDirectories.get_plugins_dir()
        if plugins_dir:
            dont_write_bytecode = sys.dont_write_bytecode
            try:
                cls._load_plugins(plugins_dir)
            finally:
                sys.dont_write_bytecode = dont_write_bytecode

    @staticmethod
    def _load_plugins(plugins_dir):
        print("[PLUGINS] Loading plugins")
        for full_name in os.listdir(plugins_dir):
            name, ext = os.path.splitext(full_name)
            if ext.lower() != ".py":
                continue
            path = os.path.join(plugins_dir, full_name)
            print("[PLUGINS] Loading", path)
            name = "plugin_" + hashlib.sha1(path.encode("UTF-8")).hexdigest()
            print(name)
            # noinspection PyDeprecation
            import imp

            try:
                # noinspection PyDeprecation
                plugin = imp.load_source(name, path)
                print("[PLUGINS] Name:", getattr(plugin, "name", ""))
                print("[PLUGINS] Version:", getattr(plugin, "version", ""))
                plugin.fsgs = fsgs
                plugin.plugin_dir = os.path.dirname(path)
                plugin.fsgs_init()
            except Exception:
                traceback.print_exc()
                continue
            else:
                print("[PLUGINS]", name, "initialized")

    settings_loaded = False

    @classmethod
    def load_settings(cls):
        if cls.settings_loaded:
            return
        cls.settings_loaded = True

        settings = Settings.instance()
        settings.load()
        path = settings.path
        # path = app.get_settings_path()
        print("loading last config from " + repr(path))
        if not os.path.exists(path):
            print("settings file does not exist")
        # noinspection PyArgumentList
        cp = ConfigParser(interpolation=None)
        try:
            cp.read([path])
        except Exception as e:
            print(repr(e))
            return

        for key in LauncherSettings.old_keys:
            if app.settings.get(key):
                print("[SETTINGS] Removing old key", key)
                app.settings.set(key, "")

        if fsgs.config.add_from_argv():
            print("[CONFIG] Configuration specified via command line")
            # Prevent the launcher from loading the last used game
            LauncherSettings.set("parent_uuid", "")
        elif LauncherSettings.get("config_path"):
            if LauncherConfig.load_file(LauncherSettings.get("config_path")):
                print("[CONFIG] Loaded last configuration file")
            else:
                print("[CONFIG] Failed to load last configuration file")
                LauncherConfig.load_default_config()
        else:
            pass
            # config = {}
            # try:
            #     keys = cp.options("config")
            # except NoSectionError:
            #     keys = []
            # for key in keys:
            #     config[key] = fs.from_utf8_str(cp.get("config", key))
            # for key, value in config.items():
            #     print("loaded", key, value)
            #     fsgs.config.values[key] = value

        # Argument --new-config[=<platform>]
        new_config = "--new-config" in sys.argv
        new_config_platform = None
        for platform_id in PLATFORM_IDS:
            if "--new-config=" + platform_id in sys.argv:
                new_config = True
                new_config_platform = platform_id
        if new_config:
            LauncherConfig.load_default_config(platform=new_config_platform)
            # Prevent the launcher from loading the last used game
            LauncherSettings.set("parent_uuid", "")

    @staticmethod
    def save_settings():
        extra = {}
        for key, value in fsgs.config.values.items():
            if key.startswith("__"):
                # keys starting with __ are never saved
                continue
            extra["config/" + str(key)] = str(value)
        app.settings.save(extra=extra)

    @classmethod
    def start_game(cls):
        from .netplay.netplay import Netplay

        if Netplay.current() and Netplay.current().game_channel:
            Netplay.current().start_netplay_game()
        else:
            cls.start_local_game()

    @classmethod
    def start_local_game(cls):
        print("START LOCAL GAME")
        print("x_missing_files", LauncherConfig.get("x_missing_files"))

        if LauncherConfig.get("x_missing_files"):
            if LauncherConfig.get("download_file"):
                if LauncherConfig.get(
                    "download_terms"
                ) and not Downloader.check_terms_accepted(
                    LauncherConfig.get("download_file"),
                    LauncherConfig.get("download_terms"),
                ):
                    from .ui.launcherwindow import LauncherWindow

                    dialog = DownloadTermsDialog(
                        LauncherWindow.current(), fsgs
                    )
                    if not dialog.show_modal():
                        return

            elif LauncherConfig.get("download_page"):
                from .ui.launcherwindow import LauncherWindow

                # fsui.show_error(_("This game must be downloaded first."))
                DownloadGameWindow(LauncherWindow.current(), fsgs).show()
                return
            else:
                fsui.show_error(
                    gettext(
                        "This game variant cannot be started "
                        "because you don't have all required files."
                    )
                )
                return

        platform_id = LauncherConfig.get(Option.PLATFORM).lower()
        if platform_id in AMIGA_PLATFORMS:
            cls.start_local_game_amiga()
        else:
            cls.start_local_game_other()

    @classmethod
    def start_local_game_other(cls):
        if True:
            platform_id = LauncherConfig.get(Option.PLATFORM).lower()
            platform_handler = PlatformHandler.create(platform_id)
        else:
            database_name = LauncherConfig.get("__database")
            variant_uuid = LauncherConfig.get("variant_uuid")
            assert variant_uuid
            fsgs.game.set_from_variant_uuid(database_name, variant_uuid)
            platform_handler = PlatformHandler.create(fsgs.game.platform.id)

        runner = platform_handler.get_runner(fsgs)
        task = RunnerTask(runner)
        from .ui.launcherwindow import LauncherWindow

        dialog = LaunchDialog(
            LauncherWindow.current(), gettext("Launching Game"), task
        )
        dialog.show()
        LauncherConfig.set("__running", "1")
        task.start()
        # dialog.show_modal()
        # dialog.close()

    @classmethod
    def start_local_game_amiga(cls):
        # make sure x_kickstart_file is initialized
        LauncherConfig.set_kickstart_from_model()

        # if not Config.get("x_kickstart_file"):  # or not \
        #     #  os.path.exists(Config.get("kickstart_file")):
        #     fsui.show_error(
        #         gettext("No kickstart found for this model. Use the 'Import "
        #                 "Kickstarts' function from the menu."))
        #     return
        cs = Amiga.get_model_config(LauncherConfig.get("amiga_model"))[
            "ext_roms"
        ]
        if len(cs) > 0:
            # extended kickstart ROM is needed
            if not LauncherConfig.get("x_kickstart_ext_file"):
                fsui.show_error(
                    gettext(
                        "No extended kickstart found for this model. "
                        "Try 'scan' function."
                    )
                )
                return

        config = LauncherConfig.copy()
        prepared_config = cls.prepare_config(config)

        model = LauncherConfig.get("amiga_model")
        if model.startswith("CD32"):
            platform = "CD32"
        elif model == "CDTV":
            platform = "CDTV"
        else:
            platform = "Amiga"
        name = LauncherSettings.get("config_name")
        uuid = LauncherConfig.get("x_game_uuid")

        from fsgs.saves import SaveHandler

        save_state_handler = SaveHandler(fsgs, name, platform, uuid)

        from fsgs.amiga.launchhandler import LaunchHandler

        launch_handler = LaunchHandler(
            fsgs, name, prepared_config, save_state_handler
        )

        from .ui.launcherwindow import LauncherWindow

        task = AmigaLaunchTask(launch_handler)
        # dialog = LaunchDialog(MainWindow.instance, launch_handler)
        dialog = LaunchDialog(
            LauncherWindow.current(), gettext("Launching FS-UAE"), task
        )
        dialog.show()

        def on_show_license_information(license_text):
            unused(license_text)
            # FIXME: don't depend on wx here
            # noinspection PyUnresolvedReferences
            # import wx
            # license_dialog = wx.MessageDialog(
            #     dialog, license_text, _("Terms of Use"),
            #     wx.OK | wx.CANCEL | wx.CENTRE)
            # license_dialog.CenterOnParent()
            # result = license_dialog.ShowModal()
            # return result == wx.ID_OK
            # FIXME
            return True

        fsgs.file.on_show_license_information = on_show_license_information

        LauncherConfig.set("__running", "1")
        task.start()
        # dialog.show_modal()
        # dialog.close()

    @classmethod
    def prepare_config(cls, original_config):
        config = StringDict(str)
        for key, value in LauncherSettings.items():
            # We now show warnings on status bar instead
            # if key in LauncherConfig.config_keys:
            #     print("... ignoring config key from settings:", key)
            #     continue
            config[key] = value

        config["base_dir"] = FSGSDirectories.get_base_dir()

        for key, value in original_config.items():
            if value:
                config[key] = value

        if not config["joystick_port_0_mode"]:
            config["joystick_port_0_mode"] = "mouse"
        if not config["joystick_port_1_mode"]:
            if config["amiga_model"].startswith("CD32"):
                config["joystick_port_1_mode"] = "cd32 gamepad"
            else:
                config["joystick_port_1_mode"] = "joystick"
        if not config["joystick_port_2_mode"]:
            config["joystick_port_2_mode"] = "none"
        if not config["joystick_port_3_mode"]:
            config["joystick_port_3_mode"] = "none"

        from .devicemanager import DeviceManager

        devices = DeviceManager.get_devices_for_ports(config)
        for port in range(4):
            key = "joystick_port_{0}".format(port)
            if not config.get(key):
                # key not set, use calculated default value
                config[key] = devices[port].id

        for remove_key in [
            "database_username",
            "database_password",
            "database_username",
            "database_email",
            "database_auth",
            "device_id",
        ]:
            if remove_key in config:
                del config[remove_key]

        # overwrite netplay config

        if config.get("__netplay_host", ""):
            config["netplay_server"] = config["__netplay_host"]
        if config.get("__netplay_password", ""):
            config["netplay_password"] = config["__netplay_password"]
        if config.get("__netplay_port", ""):
            config["netplay_port"] = config["__netplay_port"]

        # copy actual kickstart options from x_ options

        config["kickstart_file"] = config["x_kickstart_file"]
        config["kickstart_ext_file"] = config["x_kickstart_ext_file"]

        if not config["kickstart_file"]:
            # Warning will have been shown on the status bar
            config["kickstart_file"] = "internal"

        # Copy default configuration values from model defaults. The main
        # purpose of this is to let the launch code know about implied defaults
        # so it can for example configure correct ROM files for expansions.

        model_config = Amiga.get_current_config(config)
        for key, value in model_config["defaults"].items():
            if not config.get(key):
                config[key] = value

        # make sure FS-UAE does not load other config files (Host.fs-uae)
        config["end_config"] = "1"
        # Make FS-UAE check that version matches (except for development)
        # Edit: The Launcher and FS-UAE will not have identical version
        # numbers going forward.
        # if VERSION != "9.8.7dummy":
        #     config[Option.EXPECT_VERSION] = VERSION

        if config.get("__netplay_game", ""):
            print("\nfixing config for netplay game")
            for key in [x for x in config.keys() if x.startswith("uae_")]:
                print("* removing option", key)
                del config[key]

        return config


# FIXME: Files to clean up:
# Documents/FS-UAE/Cache/File Database.sqlite
# Documents/FS-UAE/Data/Game Database.sqlite
# Documents/FS-UAE/Data/Launcher.sqlite
# Documents/FS-UAE/Logs/FS-UAE.log
# Documents/FS-UAE/Logs/FS-UAE.log.txt
# Documents/FS-UAE/Data/Database.sqlite
# Documents/FS-UAE/Cache/Files.sqlite
# Documents/FS-UAE/Cache/Locker.sqlite
# Documents/FS-UAE/Cache/oagd.net.sqlite
# Documents/FS-UAE/Cache/Games.sqlite


class AmigaLaunchTask(Task):
    def __init__(self, launch_handler):
        Task.__init__(self, "Amiga Launch Task")
        self.launch_handler = launch_handler

    def run(self):
        self.launch_handler.run_sequence()


class RunnerTask(Task):
    def __init__(self, driver):
        Task.__init__(self, "RunnerTask")
        self.driver = driver

    @property
    def runner(self):
        warnings.warn("Deprecated", DeprecationWarning)
        return self.driver

    def __del__(self):
        print("RunnerTask.__del__")

    def run(self):
        device_helper = EnumerateHelper()
        device_helper.default_port_selection(
            self.driver.ports, self.driver.options
        )

        self.driver.prepare()
        self.driver.install()
        self.set_progress("__run__")
        self.driver.run()
        self.driver.wait()
        self.driver.finish()



import fsui
from launcher.i18n import gettext
from launcher.setup.setupwelcomepage import SetupWelcomePage
from launcher.ui.skin import LauncherTheme
from launcher.ui.widgets import CloseButton, NextButton, PrevButton


class SetupWizardDialog(fsui.Window):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent):
        super().__init__(
            parent,
            gettext("Setup Wizard"),
            minimizable=False,
            maximizable=False,
        )
        self.theme = LauncherTheme.get()
        self.layout = fsui.VerticalLayout()

        page = SetupWelcomePage(self)
        self.layout.add(page, expand=True, fill=True)

        button_layout = fsui.HorizontalLayout()
        self.layout.add(button_layout, fill=True, margin=20)
        button_layout.add_spacer(0, expand=True)
        self.prev_button = PrevButton(self)
        button_layout.add(self.prev_button, fill=True, margin_left=10)
        self.next_button = NextButton(self)
        button_layout.add(self.next_button, fill=True, margin_left=10)
        if self.window.theme.has_close_buttons:
            self.close_button = CloseButton(self)
            button_layout.add(self.close_button, fill=True, margin_left=10)



import os
import shutil

from fsbc.fs import get_data_dir
from fsgs.FSGSDirectories import FSGSDirectories


class ApplicationMixin(object):
    def get_settings_path(self):
        settings_ini = os.path.join(
            FSGSDirectories.get_data_dir(), "Settings.ini"
        )
        if not os.path.exists(settings_ini):
            migrate_list = [
                os.path.join(
                    FSGSDirectories.get_data_dir(), "FS-UAE Launcher.ini"
                )
            ]
            portable_ini = os.path.join(
                FSGSDirectories.get_base_dir(), "Portable.ini"
            )
            if not os.path.exists(portable_ini):
                # not portable, can migrate settings from old
                # launcher.settings file
                migrate_list.append(
                    os.path.join(get_data_dir(), "fs-uae", "launcher.settings")
                )
            # move the highest-priority settings file if present
            for migrate in migrate_list:
                if os.path.exists(migrate):
                    shutil.move(migrate, settings_ini)
                    break
            # remove all old settings files
            for migrate in migrate_list:
                if os.path.exists(migrate):
                    os.remove(migrate)

        return settings_ini



import logging

import fsui
from fsbc.application import app
from fsgs.Database import Database
from fsgs.option import Option
from launcher.i18n import gettext
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior


class GameListSelector(fsui.Choice):
    def __init__(self, parent):
        fsui.Choice.__init__(self, parent)
        self.game_lists = []
        self.populate_list()
        self.changed.connect(self.__on_changed_signal)
        SettingsBehavior(self, [Option.CONFIG_REFRESH, Option.GAME_LIST_UUID])

    def on_settings(self, key, value):
        if key == Option.CONFIG_REFRESH:
            list_found = False
            with self.changed.inhibit:
                logging.debug("GameListSelector on config_refresh")
                old_list_uuid = self.selected_list_uuid()
                logging.debug(
                    "GameListSelector old list uuid %s", repr(old_list_uuid)
                )
                logging.debug("GameListSelector Set choice index to None")
                self.set_index(None)
                self.populate_list()
                logging.debug(
                    "GameListSelector game lists: %s", repr(self.game_lists)
                )
                for i, item in enumerate(self.game_lists):
                    logging.debug("GameListSelector - %s", repr(item[0]))
                    if item[0] == old_list_uuid:
                        if self.get_index() != i:
                            logging.debug(
                                "GameListSelector Set choice index to %d", i
                            )
                            self.set_index(i)
                        list_found = True
                        break
            if not list_found:
                logging.debug("GameListSelector list uuid is no longer valid")
                self.set_index(0)
        elif key == Option.GAME_LIST_UUID:
            logging.debug("GameListSelector GAME LIST UUID %s", value)
            self.select_list_from_uuid(value)

    def select_list_from_uuid(self, uuid):
        for i, item in enumerate(self.game_lists):
            if item[0] == uuid:
                if self.get_index() != i:
                    with self.changed.inhibit:
                        logging.debug("GameListSelector = %d", i)
                        self.set_index(i)
                break
        else:
            self.set_index(0)

    def populate_list(self):
        database = Database.instance()
        self.game_lists = database.get_game_lists()
        if len(self.game_lists) > 0:
            self.game_lists.insert(0, ["", self.ITEM_SEPARATOR])
        self.game_lists.insert(0, [Database.GAME_LIST_GAMES, gettext("Games")])
        self.game_lists.insert(
            0, [Database.GAME_LIST_CONFIGS, gettext("Configs")]
        )
        self.game_lists.insert(0, ["", gettext("Configs and Games")])
        self.clear()
        for item in self.game_lists:
            list_name = item[1]
            if list_name == "Favorites":
                list_name = gettext("Favorites")
            self.add_item(list_name)

    def selected_list_uuid(self):
        index = self.get_index()
        try:
            return self.game_lists[index][0]
        except IndexError:
            # This should not really happen...
            return ""

    def __on_changed_signal(self):
        app.settings["game_list_uuid"] = self.selected_list_uuid()



from fsgs.platform import PlatformHandler
from fsgs.platforms.a2600.messa2600driver import MessA2600Driver
from fsgs.platforms.loader import SimpleLoader


class Atari2600PlatformHandler(PlatformHandler):
    PLATFORM_NAME = "Atari 2600"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return Atari2600Loader(fsgs)

    def get_runner(self, fsgs):
        return MessA2600Driver(fsgs)


class Atari2600Loader(SimpleLoader):
    pass



import weakref

from launcher.launcher_config import LauncherConfig


class ConfigBehavior:
    def __init__(self, parent, names):
        parent.__config_enable_behavior = self
        self._parent = weakref.ref(parent)
        self._names = set(names)
        LauncherConfig.add_listener(self)
        try:
            parent.destroyed.connect(self.on_parent_destroyed)
        except AttributeError:
            print(
                "WARNING: ConfigBehavior without remove_listener "
                "implementation"
            )
        for name in names:
            # Broadcast initial value
            self.on_config(name, LauncherConfig.get(name))

    def on_parent_destroyed(self):
        print("ConfigBehavior: remove_listener", self._parent())
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key in self._names:
            widget = self._parent()
            func = getattr(widget, "on_{0}_config".format(key))
            func(value)



import time

from fsbc.application import app
from fsbc.signal import Signal
from fsbc.task import Task
from fsgs.Database import Database
from fsgs.ogd.context import SynchronizerContext
from fsgs.ogd.lists import ListsSynchronizer
from fsgs.ogd.meta import MetaSynchronizer


class DatabaseRefreshTask(Task):
    def __init__(self):
        Task.__init__(self, "DatabaseRefreshTask")

    # def stop_check(self):
    #     pass

    def on_status(self, status):
        self.progressed(status)

    def run(self):
        with Database.get_instance() as database:
            self._run(database)

    def _run(self, database):
        # FIXME, dependency on fs_uae_launcher
        # from fs_uae_launcher.Scanner import Scanner
        # Scanner.start([], scan_for_files=False, update_game_database=True)
        from fsgs.ogd.game_rating_synchronizer import GameRatingSynchronizer
        from launcher.gamescanner import GameScanner

        context = SynchronizerContext()

        synchronizer = MetaSynchronizer(
            context, on_status=self.on_status, stop_check=self.stop_check
        )
        synchronizer.synchronize()

        synchronizer = GameRatingSynchronizer(
            context,
            database,
            on_status=self.on_status,
            stop_check=self.stop_check,
        )
        synchronizer.username = "auth_token"
        synchronizer.password = app.settings["database_auth"]
        synchronizer.synchronize()

        synchronizer = ListsSynchronizer(
            context, on_status=self.on_status, stop_check=self.stop_check
        )
        synchronizer.synchronize()

        scanner = GameScanner(
            context, None, on_status=self.on_status, stop_check=self.stop_check
        )
        scanner.update_game_database()
        scanner.scan(database)

        # FIXME: review what signals should be sent when a scan is performed
        # FIXME: these should be removed soon
        app.settings["last_scan"] = str(time.time())
        app.settings["__config_refresh"] = str(time.time())
        # this must be called from main, since callbacks are broadcast
        # when settings are changed
        Signal("scan_done").notify()



import fsui
from fsui.qt import QObject, QSignal

from .application import Application as Application
from .button import Button as Button
from .button import CustomButton as CustomButton
from .button import FlatButton as FlatButton
from .canvas import Canvas as Canvas
from .color import Color as Color
from .font import Font as Font
from .image import Image as Image
from .image_button import ImageButton as ImageButton
from .image_view import ImageView as ImageView
from .label import Label as Label
from .label import MultiLineLabel as MultiLineLabel
from .layout import Column as Column
from .layout import Row as Row
from .menu import Menu as Menu
from .painter import Painter as Painter
from .panel import Panel as Panel
from .text_area import TextArea as TextArea
from .text_field import TextField as TextField
from .window import TitlePanel as TitlePanel
from .window import TitleSeparator as TitleSeparator
from .window import Window as Window


class Object(QObject):
    pass


# noinspection PyPep8Naming
def Signal():
    return QSignal()


def screen_size():
    return fsui.get_screen_size()



from __future__ import absolute_import
from __future__ import print_function

from .ADFSDir import ADFSDir
from .MetaInfo import MetaInfo
from . import DosType

class ADFSVolDir(ADFSDir):
  def __init__(self, volume, root_block):
    ADFSDir.__init__(self, volume, None)
    self.set_block(root_block)
    self._init_name_hash()
  
  def __repr__(self):
    return "[VolDir(%d)'%s':%s]" % (self.block.blk_num, self.block.name, self.entries)
  
  def draw_on_bitmap(self, bm, show_all=False, first=True):
    blk_num = self.block.blk_num
    bm[blk_num] = 'V'
    if show_all or first:
      self.ensure_entries()
      for e in self.entries:
        e.draw_on_bitmap(bm, show_all, False)

  def get_size_str(self):
    return "VOLUME"

  def create_meta_info(self):
    self.meta_info = MetaInfo(mod_ts=self.block.mod_ts)
  
  def can_delete(self):
    return False

  def get_list_str(self, indent=0, all=False, detail=False):
    a = ADFSDir.get_list_str(self, indent=indent, all=all, detail=detail)
    a += DosType.get_dos_type_str(self.volume.get_dos_type())
    return a




import json
from collections import defaultdict

from fsgs.network import openretro_url_prefix


class SimpleLoader:
    def __init__(self, fsgs):
        self.fsgs = fsgs
        self.config = {}

    def get_config(self):
        return self.config.copy()

    def load_files(self, values):
        file_list = json.loads(values["file_list"])
        if len(file_list) == 0:
            self.config["x_variant_error"] = "Variant has empty file list"
        elif len(file_list) > 1:
            self.config["x_variant_error"] = "Unsupported multi-file variant"

        self.config["cartridge_slot"] = "sha1://{0}/{1}".format(
            file_list[0]["sha1"], file_list[0]["name"]
        )

        self.config["cue_sheets"] = values["cue_sheets"]

    def load_extra(self, values):
        pass

    def load_basic(self, values):
        self.config["command"] = values["command"]
        self.config["game_name"] = values["game_name"]
        self.config["game_uuid"] = values["game_uuid"]
        self.config["variant_name"] = values["variant_name"]
        self.config["variant_uuid"] = values["variant_uuid"]
        self.config["model"] = values["model"]
        self.config["platform"] = values["platform"]
        self.config["protection"] = values["protection"]
        self.config["viewport"] = values["viewport"]
        self.config["video_standard"] = values["video_standard"]

    def load_info(self, values):
        self.config["languages"] = values["languages"]
        self.config["players"] = values["players"]
        self.config["year"] = values["year"]
        self.config["publisher"] = values["publisher"]
        self.config["developer"] = values["developer"]

        self.config["x_game_notice"] = values["game_notice"]
        self.config["x_variant_notice"] = values["variant_notice"]
        self.config["x_variant_warning"] = values["variant_warning"]
        self.config["x_variant_error"] = values["variant_error"]

        self.config["database_url"] = "{0}/game/{1}".format(
            openretro_url_prefix(), values["parent_uuid"]
        )
        for key in ["mobygames_url"]:
            self.config[key] = values[key]

        for key in [
            "download_file",
            "download_page",
            "download_terms",
            "download_notice",
        ]:
            if key in values:
                self.config[key] = values[key]

    def load_images(self, values):
        for key in [
            "front_sha1",
            "screen1_sha1",
            "screen2_sha1",
            "screen3_sha1",
            "screen4_sha1",
            "screen5_sha1",
            "title_sha1",
        ]:
            self.config[key] = values[key]

    def load_values(self, key_values):
        # print(key_values)
        values = defaultdict(str)
        values.update(key_values)
        self.load(values)
        return self.get_config()

    def load(self, values):
        self.load_basic(values)
        self.load_extra(values)
        self.load_info(values)
        self.load_images(values)
        self.load_files(values)


class PlatformLoader(SimpleLoader):
    pass


class CartridgePlatformLoader(PlatformLoader):
    pass


class CDPlatformLoader(PlatformLoader):
    def load_files(self, values):
        self.config["cue_sheets"] = values["cue_sheets"]
        self.config["file_list"] = values["file_list"]



import fsui
from fsgs.amiga.amiga import Amiga
from fsgs.option import Option
from fsgs.platform import Platform
from fsgs.platforms.commodore64 import C64_MODEL_C64C_1541_II
from fsgs.platforms.zxs.spectrumplatform import ZXS_MODEL_PLUS3
from launcher.launcher_config import LauncherConfig
from launcher.ui.behaviors.configbehavior import ConfigBehavior
from launcher.ui.behaviors.platformbehavior import AMIGA_PLATFORMS
from launcher.ui.floppiesgroup import FloppiesGroup


class RemovableMediaGroup(FloppiesGroup):
    def __init__(self, parent, drives, main=False):
        FloppiesGroup.__init__(self, parent, drives, removable_media=True)
        self.layout3 = fsui.HorizontalLayout()
        self.layout.add(self.layout3, fill=True)
        self.layout3.add_spacer(0, expand=True)
        self.cd_mode = False
        self.__platform = ""
        self.__amiga_model = ""
        self._main = main
        self._c64_model = ""
        self._zxs_model = ""

        self._ines_header_widget = INesHeaderWidget(self)
        self._ines_header_widget.hide()
        self.layout.add(self._ines_header_widget, fill=True)
        self._a78_header_widget = A78HeaderWidget(self)
        self._a78_header_widget.hide()
        self.layout.add(self._a78_header_widget, fill=True)

        self.update_media_type()

        ConfigBehavior(
            self,
            [
                Option.PLATFORM,
                Option.AMIGA_MODEL,
                Option.C64_MODEL,
                Option.ZXS_MODEL,
            ],
        )

    def on_platform_config(self, value):
        self.__platform = value
        self.update_media_type()

    def on_amiga_model_config(self, value):
        self.__amiga_model = value
        self.update_media_type()

    def on_c64_model_config(self, value):
        self._c64_model = value
        self.update_media_type()

    def on_zxs_model_config(self, value):
        self._zxs_model = value
        self.update_media_type()

    def update_media_type(self):
        if self.__platform in AMIGA_PLATFORMS:
            self.set_cd_mode(Amiga.is_cd_based(LauncherConfig))
        elif self.__platform in [Platform.ATARI]:
            self.set_mode(self.FLOPPY_MODE)
        elif self.__platform in [Platform.C64]:
            if self._c64_model == C64_MODEL_C64C_1541_II:
                self.set_mode(self.FLOPPY_MODE)
            else:
                self.set_mode(self.TAPE_MODE)
        elif self.__platform in [Platform.CPC]:
            self.set_mode(self.TAPE_MODE)
        elif self.__platform in [Platform.DOS]:
            self.set_mode(self.FLOPPY_MODE)
        elif self.__platform in [Platform.PSX]:
            self.set_mode(self.CD_MODE)
        elif self.__platform in [Platform.ZXS]:
            if self._zxs_model == ZXS_MODEL_PLUS3:
                self.set_mode(self.FLOPPY_MODE)
            else:
                self.set_mode(self.TAPE_MODE)
        else:
            self.set_mode(self.CARTRIDGE_MODE)

        if self._main:
            if self.__platform == Platform.A7800:
                self.selectors[1].hide()
                self._a78_header_widget.show()
                self._ines_header_widget.hide()
            elif self.__platform == Platform.NES:
                # if self.selectors[1].is_visible():
                self.selectors[1].hide()
                self._a78_header_widget.hide()
                self._ines_header_widget.show()
            else:
                # if not self.selectors[1].is_visible():
                self.selectors[1].show()
                self._a78_header_widget.hide()
                self._ines_header_widget.hide()
            self.layout.update()

    def set_mode(self, mode):
        if self.mode == mode:
            return
        self.mode = mode
        self.cd_mode = mode == self.CD_MODE
        for selector in self.selectors:
            # selector.set_cd_mode(self.cd_mode)
            selector.set_mode(self.mode)
        self.update_heading_label()
        # self.selectors[1].enable(not self.cd_mode)

    def set_cd_mode(self, cd_mode):
        if cd_mode:
            self.set_mode(self.CD_MODE)
        else:
            self.set_mode(self.FLOPPY_MODE)


class INesHeaderWidget(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin=10)

        self.text_field = fsui.TextField(self, "")
        self.text_field.on_changed = self.on_text_changed
        self.text_field.disable()
        hori_layout.add(self.text_field, expand=True)

        # self.help_button = HelpButton(
        #     self, "https://fs-uae.net/docs/options/nes-ines-header")
        # hori_layout.add(self.help_button, margin_left=10)

        ConfigBehavior(self, [Option.NES_INES_HEADER])

    def on_nes_ines_header_config(self, value):
        if value != self.text_field.get_text():
            value = "iNES Header: " + value
            self.text_field.set_text(value)

    def on_text_changed(self):
        # LauncherConfig.set("nes_ines_header", self.text_field.get_text())
        pass


class A78HeaderWidget(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin=10)

        self.text_field = fsui.TextField(self, "")
        self.text_field.on_changed = self.on_text_changed
        self.text_field.disable()
        hori_layout.add(self.text_field, expand=True)

        # self.help_button = HelpButton(
        #     self, "https://fs-uae.net/docs/options/nes-ines-header")
        # hori_layout.add(self.help_button, margin_left=10)

        ConfigBehavior(self, [Option.A7800_A78_HEADER])

    def on_a7800_a78_header_config(self, value):
        if value != self.text_field.get_text():
            value = "A78 Header: " + value
            self.text_field.set_text(value)

    def on_text_changed(self):
        # LauncherConfig.set("nes_ines_header", self.text_field.get_text())
        pass



import fsui
from fsgs.option import Option
from launcher.i18n import gettext
from launcher.settings.scan_paths_group import ScanPathsGroup
from launcher.settings.settings_page import SettingsPage


class ScanSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("indexing-settings", "pkg:workspace")
        gettext("File Database Settings")
        title = gettext("File Database")
        subtitle = gettext(
            "Choose what folders you want to scan for Amiga files"
        )
        self.add_header(icon, title, subtitle)

        self.layout.add(
            fsui.MultiLineLabel(
                self,
                gettext(
                    "Choose what folders you want to scan for Amiga files"
                ),
            ),
            fill=True,
            margin_bottom=10,
        )
        self.scan_paths_group = ScanPathsGroup(self)
        self.layout.add(self.scan_paths_group, fill=True, expand=True)

        # For reset to defaults function
        self.options_on_page.add(Option.SEARCH_PATH)



"""tools to convert byte sizes into strings suitable for disk/file sizes"""

KIB_UNIT = 1024

# map unit extension to power of 1000/KIB_UNIT
scale_map = {
  'k' : 1,
  'm' : 2,
  'g' : 3,
  't' : 4
}

def to_byte_size_str(size, kibi_units=True):
  """convert a byte value into a 5 letter string"""
  if kibi_units:
    unit = KIB_UNIT
    marker = 'i'
  else:
    unit = 1000
    marker = ''
  if size < 1000:
    return "%3dB%s" % (size, marker)
  else:
    # run through scales
    for scale in "KMGT":
      next = size / unit
      if next < 10:
        frac = float(size) / unit
        if frac > 9.9:
          frac = 9.9
        return "%3.1f%s%s" % (frac, scale, marker)
      elif next < 1000:
        return "%3d%s%s" % (next, scale, marker)
      size = next
    return "NaNB%s" % marker

def parse_byte_size_str(s):
  """parse a string to derive a byte value.
     can read e.g. 10Ki, 2.1k or 2048.
     returns None if the string is invalid or a byte value
  """
  s = s.lower()
  n = len(s)
  if n == 0:
    return None
  # 'i' - use Kibi units
  if s[-1] == 'i':
    unit = KIB_UNIT
    if n == 1:
      return None
    s = s[:-1]
    n -= 1
  # else Si units
  else:
    unit = 1000
  # check for scale
  scale = s[-1]
  if scale in scale_map.keys():
    factor = unit ** scale_map[scale]
    if n == 1:
      return None
    s = s[:-1]
    n -= 1
  else:
    factor = 1
  # get number
  try:
    v = int(float(s) * factor)
    return v
  except ValueError:
    return None

# a small test
if __name__ == '__main__':
  import sys
  for a in sys.argv[1:]:
    v = parse_byte_size_str(a)
    if v != None:
      print(a, ":", v, "=", to_byte_size_str(v), "=", to_byte_size_str(v, False))
    else:
      print(a, ":", v)
      
